/* support for converting grids */

#include "headers.h"

void make_grid_cv(int ncid, struct dment *de, double start, double end, int convertDMS);

void move_grid_attributes(int gdhandle, char *gdname, int ncid);

/* a little table to map UTM zone codes (courtesy of GDprojinfo) to
   strings suitable for hanging as properties on a grid.  I was tempted
   to write a function that encodes this, but only for a moment...
*/
char *utmzonecodemap[] =
{"",
 "177W","171W","165W","159W","153W","147W","141W","135W","129W","123W",
 "117W","111W","105W","099W","093W","087W","081W","075W","069W","063W",
 "057W","051W","045W","039W","033W","027W","021W","015W","009W","003W",
 "003E","009E","015E","021E","027E","033E","039E","045E","051E","057E",
 "063E","069E","075E","081E","087E","093E","099E","105E","111E","117E",
 "123E","129E","135E","141E","147E","153E","159E","165E","171E","177E"
};

/* a table for names of GCTP Spheroid codes */
char *spheroid_names[] = {
  "Clarke 1866", "Clarke 1880", "Bessel", "International 1967",
  "International 1909", "WGS 72", "Everest", "WGS 66", "GRS 1980",
  "Airy", "Modified Airy", "Modified Everest", "WGS 84", "Southeast Asia",
  "Austrailian National", "Krassovsky", "Hough", "Mercury 1960",
  "Modified Mercury 1968", "Sphere of radius 637099 meters"
};

/* A little system for decoding the projection info (from GDprojinfo) */
struct tagval {
    int index;
    char *name;
  };
struct proj_info {
  int index;
  char *projection_name;
  int tagcount;
  struct tagval *tags;
};
/* this external declaration of the tagvals is necessary because of C's
   brain-damaged initialization semantics.  Structures can contain *foo
   pointers, but they are intepreted as pointers to exactly one foo
   (scalars).  It would be nice if you could put foo[] into a structure
   and mean you were pointing at a vector of foo.  Then we could create a
   proj_info by saying
   {1, "UTM", 2, {{0,"Longitude_Point"},{1, "Latitude_Point"}}}
   instead of the mess below.
*/

struct tagval tvUTM[] =
{{0,"Longitude_Point"},{1, "Latitude_Point"}};

struct tagval tvLCC[] =
{{0, "Semi_Major_Axis"}, {1, "Semi_Minor_Axis"},
 {2, "Latitude_First_Standard_Parallel"}, {3, "Latitude_Second_Standard_Parallel"},
 {4, "Longitude_Central_Meridian"}, {5, "Latitude_Origin"},
 {6, "False_Easting"}, {7, "False_Northing"}};

struct tagval tvPS[] =
{{0, "Semi_Major_Axis"}, {1, "Semi_Minor_Axis"},
 {4, "Longitude_Below_Pole"}, {5, "Latitude_True_Scale"},
 {6, "False_Easting"}, {7, "False_Northing"}};

struct tagval tvPC[] =
{{0, "Semi_Major_Axis"}, {1, "Semi_Minor_Axis"},
 {4, "Longitude_Central_Meridian"}, {5, "Latitude_Origin"},
 {6, "False_Easting"}, {7, "False_Northing"}};

struct tagval tvTM[] =
{{0, "Semi_Major_Axis"}, {1, "Semi_Minor_Axis"},
 {2, "Central_Meridian_Scale_Factor"},
 {4, "Longitude_Central_Meridian"}, {5, "Latitude_Origin"},
 {6, "False_Easting"}, {7, "False_Northing"}};

struct tagval tvLA[] =
{{0, "Reference_Sphere_Radius"},
 {4, "Longitude_Center"}, {5, "Latitude_Center"},
 {6, "False_Easting"}, {7, "False_Northing"}};

struct tagval tvHOM[] =
{{0, "Semi_Major_Axis"}, {1, "Semi_Minor_Axis"},
 {2, "Central_Meridian_Scale_Factor"}, {3, "Azimuth_Angle"},
 {4, "Azimuth_Longitude"}, {5, "Latitude_Origin"},
 {6, "False_Easting"}, {7, "False_Northing"},
 {8, "Longitude_Center_Line1"}, {9, "Latitude_Center_Line1"},
 {10, "Longitude_Center_Line2"}, {11, "Latitude_Center_Line2"},
 {12, "format_0A_1B"}};

struct tagval tvSOM[] =
{{0, "Semi_Major_Axis"}, {1, "Semi_Minor_Axis"},
 {2, "Landsat_Satellite_Number"},
 {3, "Orbit_Inclination_or_Landsat_Path_Number"},
 {4, "Longitude_Ascending_Orbit"},
 {6, "False_Easting"}, {7, "False_Northing"},
 {8, "Period_of_Revolution"},
 {9, "Satellite_Ratio"},
 {10, "Landsat_End_Of_Path"},
 {11, "HDF-EOS"},
 {12, "format_0A_1B"}};

struct tagval tvIG[] =
{{0, "Reference_Sphere_Radius"}};

struct tagval tvIS[] =
{{0, "Reference_Sphere_Radius"}, {4, "Longitude_Central_Meridian"},
 {6, "False_Easting"}, {7, "False_Northing"},
 {8, "Latitudinal_Zones"},
 {10, "Right_Justified_Columns"}};

/* This computes how many entries there are in a given tagval vector */
/* Actually, it can compute the compile-time size of any array structure */

#define tvSIZE(x) (sizeof(x)/sizeof(x[0]))

/* And this removes some redundancy in the GPI table entries */

#define GPIentry(index, label, tagvec) {index, label, tvSIZE(tagvec), tagvec}

/* finally, the main geographic projection information table */

struct proj_info GPI[] = {
  {0, "Geographic", 0, (struct tagval *) NULL},
  GPIentry(1, "Universal Transverse Mercator", tvUTM),
  GPIentry(4, "Lambert Conformal Conic", tvLCC),
  GPIentry(6, "Polar Stereographic", tvPS),
  GPIentry(7, "Polyconic", tvPC),
  GPIentry(9, "Transverse Mercator", tvTM),
  GPIentry(11, "Lambert Azimuthal", tvLA),
  GPIentry(20, "Hotin Oblique Mercator A/B", tvHOM),
  GPIentry(22, "Space Oblique Mercator A/B", tvSOM),
  GPIentry(24, "Interrupted Goode Homolosine", tvIG),
  GPIentry(99, "Integerized Sinusoidal Projection", tvIS)
};

struct proj_info *find_GPI(int index)
{
  int i;
  for (i=0; i< (sizeof(GPI)/sizeof(struct proj_info)); i++)
    if(GPI[i].index == index) return &(GPI[i]);
  return NULL;
}

void convert_grid(char *hdfname, char *netcdfname, int ncid)
{
  /* This is almost the same as convert_swath, except:
     - There aren't any geolocation arrays in the grid file.
     - There *is* reference point and upleft/lowright information
     that should be unrolled into netCDF coordinate vectors.
  */
  /* One more thing - apparently grids have two magic special-case
     dimensions named "XDim" and "YDim".  GDinqdims will not report these
     dimensions, but GDfieldinfo will.  We use the former to walk the
     dimensions and find any that might be Unlimited, we use the latter
     to get the real current size of dimensions for data fetch purposes.
     Getting around this requires weird special-case code which will be
     flagged with XDIM/YDIM in neighboring comments.  Bleah.

     It's actually worse than that - XDim and YDim are unique in that
     they can be DIFFERENT SIZES in different fields - they are not
     global across a grid like the dimensions reported by GDinqdims.
     This not only requires special-case code, but giving these dimensions
     unique-per-field names, which likely breaks COADS compatibility.  It's
     either that or generate multiple netCDF files per grid file....
  */

  char *gdnamebuf;
  char *gdname;
  hid_t filehandle;
  int ngrid;
  int gdindex;
  struct dment *dmap = NULL;
  
  /* a structure for remembering the origin/size/shape of grids,
     so we can complain if grids aren't congruent */
  struct gridshape {
    long xsize;
    long ysize;
    double upleft[2];
    double lowright[2];
  };
  /* the placew we remember the current shape and names
     of XDim and YDim */
  struct gridshape ourshape = {0,0,{0.0,0.0},{0.0,0.0}};
  char *xdimname = "";
  char *ydimname = "";

  {
    /* get the number of grids and their names */
    long strbufsize;
    SS(ngrid = HE5_GDinqgrid(hdfname, NULL, &strbufsize));
    gdnamebuf = newstring(strbufsize + 1);
    SS(ngrid = HE5_GDinqgrid(hdfname, gdnamebuf, &strbufsize));
  }

  SS(filehandle = HE5_GDopen(hdfname, H5F_ACC_RDONLY));

  TRACE((stderr, "File contains grid(s):\n\t|%s|\n", gdnamebuf));
  gdname = gdnamebuf;
  for(gdindex=0; gdindex<ngrid; gdindex++) {
    hid_t gdhandle;
    int numdim;
    long bufsize;
    char *dimnamebuf;
    char *dimname;
    hsize_t *dimsizebuf;
    int fieldindex;
    int numfields;
    char *fieldnamebuf = newstring(HE5_HDFE_NAMBUFSIZE);
    char *fieldname;
    int *dimidbuf;
    hsize_t *startdim;
    hsize_t *edgedim;
    int totaldatasize;
    int i;
    int projcode, zonecode, spherecode;
    double projparm[13];


    end_name(gdname);

    /* grab the grid's handle */
    TRACE((stderr, "Attaching to grid |%s|\n", gdname));
    SS(gdhandle = HE5_GDattach(filehandle, gdname));

    /* unroll its dimension names and sizes */
    SS(numdim = HE5_GDnentries(gdhandle, HE5_HDFE_NENTDIM, &bufsize));
    /* This API call can sometimes report that a grid has NO dimensions,
       if, say, it only has XDim and YDim. So, we have to check for
       the zero case below.
    */
    /* XDIM/YDIM - we allocate these buffers just big enough to handle
       the dimensions found by GDinqdims, plus XDim and YDim
    */
    dimnamebuf = newstring(bufsize + 1 + 10);
    dimsizebuf = (hsize_t *) malloc(sizeof(hsize_t) * (numdim+2));

    if(numdim > 0) {
      SS(numdim = HE5_GDinqdims(gdhandle, dimnamebuf, dimsizebuf));
      TRACE((stderr, "%d dimensions:\n\t|%s|\n", numdim, dimnamebuf));

      /* For each dimension, make an entry in the dimension map */
      dimname = dimnamebuf;
      for(i=0; i<numdim; i++) {
	struct dment *olddim = find_dimension(dmap, dimname);
	end_name(dimname);
	if(olddim == NULL) {
	  int newdimid;
	  TRACE((stderr, "defining dimension |%s|, size %d\n",
		 dimname, dimsizebuf[i]));
	  /* Map HDF Unlimited to netCDF unlimited here */
	  if(dimsizebuf[i] == H5S_UNLIMITED)
	    dimsizebuf[i] = NC_UNLIMITED;
	  nc_redef(ncid);
	  SS(nc_def_dim(ncid, dimname, dimsizebuf[i], &newdimid));
	  add_dimension(&dmap, dimname, dimname, dimsizebuf[i], newdimid);
	}
	else {
	  /* If we've seen this name before, but with a different size,
	     smash the old entry with a new netCDF name and size
	  */
	  if(olddim->size != dimsizebuf[i]) {
	    char *newdimname;
	    TRACE((stderr, "Dimension %s has old size %d and new size %d)!\n",
		   dimname, olddim->size, dimsizebuf[i]));
	    newdimname = newstring(strlen(dimname) + strlen(gdname) + 2);
	    sprintf(newdimname, "%s_%s", dimname, gdname);
	    TRACE((stderr, "Creating new netCDF dimension %s\n", newdimname));
	    if(olddim->Nname != olddim->Hname)
	      free(olddim->Nname);
	    olddim->Nname = newdimname;
	    olddim->size = dimsizebuf[i];
	    nc_redef(ncid);
	    SS(nc_def_dim(ncid, dimname, dimsizebuf[i], &(olddim->id)));
	  }
	}
	dimname = next_name(dimname);
      }
    }


    /* check the grid's shape and geolocation information.
       We need to insure that XDim and YDim are in the dimension map,
       that they have netCDF definitions (with useful names),
       and their coordinate vectors are properly filled in.
    */
    {
      struct gridshape current;
      struct dment *xentry;
      struct dment *yentry;
      char *xname;
      char *yname;
      int newdimid;
      int convertDMS = 0;

      SS(HE5_GDgridinfo(gdhandle, &(current.xsize), &(current.ysize),
			current.upleft, current.lowright));
      SS(HE5_GDprojinfo(gdhandle, &projcode, &zonecode, &spherecode,
			projparm));
      TRACE((stderr, "Grid info: size (%d,%d) upleft (%f, %f) lowright (%f, %f)\n",
	     current.xsize, current.ysize,
	     current.upleft[0], current.upleft[1],
	     current.lowright[0], current.lowright[1]));
      TRACE((stderr, "Projection info: code %d, zonecode %d spherecode %d\n",
	     projcode, zonecode, spherecode));
      /* According to one of the programmers for HDF-EOS, only projection
	 HE5_GCTP_GEO (geographic) is longitude/latitude.  All of the other
	 grid projections are UTMish, and should have their coordinates
	 named easting/northing
      */
      switch(projcode) {
      case HE5_GCTP_GEO:
	xname = "longitude";
	yname = "latitude";
	convertDMS = 1;
	break;
      default: /* UTM */
	xname = "easting";
	yname = "northing";
	break;
      }
      if(ourshape.xsize == 0 && ourshape.ysize == 0) {
	/* this is the first one - initialize ourshape */
	ourshape = current;
	nc_redef(ncid);
	SS(nc_def_dim(ncid, xname, current.xsize, &newdimid));
	xentry = add_dimension(&dmap, "XDim", xname, current.xsize, newdimid);

	SS(nc_def_dim(ncid, yname, current.ysize, &newdimid));
	yentry = add_dimension(&dmap, "YDim", yname, current.ysize, newdimid);

	make_grid_cv(ncid, xentry,
		     current.upleft[0], current.lowright[0], convertDMS);
	make_grid_cv(ncid, yentry,
		     current.upleft[1], current.lowright[1], convertDMS);
      }
      else {
	/* check the XDim - if it differs from the one last time, update it */
	xentry = find_dimension(dmap, "XDim");
	if (!(ourshape.xsize == current.xsize
	      && ourshape.upleft[0] == current.upleft[0]
	      && ourshape.lowright[0] == current.lowright[0]
	      && !strcmp(xname, xentry->Nname))) {
	  char *newdimname = newstring(strlen(xname) + strlen(gdname) + 2);
	  
	  ourshape.xsize = current.xsize;
	  ourshape.upleft[0] = current.upleft[0];
	  ourshape.lowright[0] = current.lowright[0];
	  sprintf(newdimname, "%s_%s", gdname, xname);
	  nc_redef(ncid);
	  SS(nc_def_dim(ncid, newdimname, current.xsize, &newdimid));
	  /* this assignment can cause a little, bitty memory leak */
	  xentry->Nname = newdimname;
	  xentry->size = current.xsize;
	  xentry->id = newdimid;

	  make_grid_cv(ncid, xentry,
		       current.upleft[0], current.lowright[0], convertDMS);
	}

	/* check the YDim - if it differs from the one last time, update it */
	yentry = find_dimension(dmap, "YDim");

	if (!(ourshape.ysize == current.ysize
	      && ourshape.upleft[1] == current.upleft[1]
	      && ourshape.lowright[1] == current.lowright[1]
	      && !strcmp(yname, yentry->Nname))) {
	  char *newdimname = newstring(strlen(yname) + strlen(gdname) + 2);

	  ourshape.ysize = current.ysize;
	  ourshape.upleft[1] = current.upleft[1];
	  ourshape.lowright[1] = current.lowright[1];
	  sprintf(newdimname, "%s_%s", gdname, yname);
	  nc_redef(ncid);
	  SS(nc_def_dim(ncid, newdimname, current.ysize, &newdimid));
	  /* this assignment can cause a little, bitty memory leak */
	  yentry->Nname = newdimname;
	  yentry->size = current.ysize;
	  yentry->id = newdimid;

	  make_grid_cv(ncid, yentry,
		       current.upleft[1], current.lowright[1], convertDMS);
	}
      }
    }


    /* the dimension map is now filled - hose the fields into the netCDF file
    */

   
    /* unroll its field names and data types */
    SS(numfields = HE5_GDnentries(gdhandle, HE5_HDFE_NENTDFLD, &bufsize));
    if(numfields > 0) {
      fieldnamebuf = newstring(bufsize + 1);
      SS(numfields = HE5_GDinqfields(gdhandle, fieldnamebuf, NULL, NULL));

      TRACE((stderr, "Attached to grid |%s| with %d fields:\n\t|%s|\n",
	     gdname, numfields, fieldnamebuf));

      /* OK, each field then maps to a variable name (grid_varname, to
	 insure uniqueness, I guess) in the netCDF file */
      fieldname = fieldnamebuf;
      for(fieldindex=0; fieldindex< numfields; fieldindex++) {
	int fieldtype;
	long totalfieldentries = 1;
	void *databuf;
	int varid;
	char *varname;
	int nctype;

	end_name(fieldname);

	/* This is how to determine the real type of a field,
	   the number of dimensions it has,
	   and the real size of each dimension.  We can safely clobber
	   the existing dimsizebuf and dimnamebuf because the dimensions for
	   a field have to be a subset of the dimensions for its grid
	   (if this isn't true, some HDF-EOS designer needs a swat with a
	   clue-by-four). XDIM/YDIM, dammit!
	*/
	SS(HE5_GDfieldinfo(gdhandle, fieldname,
			   &numdim, dimsizebuf, &fieldtype, dimnamebuf, NULL));
	dimidbuf = (int *) malloc(numdim * sizeof(int));
	/* as long as we're here, allocate some space for the readfield
	   arrays...
	*/
	startdim = (hsize_t *) malloc(numdim * sizeof(hsize_t));
	/* edgedim = (hsize_t *) malloc(numdim * sizeof(hsize_t)); */

	TRACE((stderr, "Processing field %s with %d dimensions:\n\t|%s|\n",
	       fieldname, numdim, dimnamebuf));

	/* compute the total number of elements in this grid */
	/* and build a dimension ID vector for variable creation later */
	dimname = dimnamebuf;
	for(i=0; i<numdim; i++) {
	  struct dment *dimentry;
	  end_name(dimname);
	  dimentry = find_dimension(dmap, dimname);
	  if(dimentry == NULL) {
	    /* A new (to us) dimension - map it.
	       There is apparently no reliable way to determine if
	       this dimension is unlimited - this API call returns
	       the actual size.  So, if this dimension was unlimited,
	       we won't record that in the netCDF file.  Pity...
	    */
	    int newdimid;
	    TRACE((stderr, "defining late dimension |%s|, size %d\n",
		   dimname, dimsizebuf[i]));
	    SS(nc_def_dim(ncid, dimname, dimsizebuf[i], &newdimid));
	    dimentry = add_dimension(&dmap, dimname, dimname, dimsizebuf[i], newdimid);
	  }
	  else {
	    /* If this dimension doesn't match what we looked up, we
	       can't translate this file without serious conniptions
	       (changing the name of the dimension, creating multiple
	       netCDF files), so we error out if that happens
	    */
	    if(dimentry->size != dimsizebuf[i]) {
	      fprintf(stderr, "Dimension %s has two different sizes (%d and %d)!\n",
		      dimname, dimentry->size, dimsizebuf[i]);
	      spit_status(-1, "bad dimension sizes");
	    }
	  }
	  TRACE((stderr, "Found dimension |%s|, id %d, size %d\n",
		 dimname, dimentry->id, dimentry->size));
	  dimidbuf[i] = dimentry->id;
	  totalfieldentries *= dimsizebuf[i];
	  startdim[i] = 0;

	  dimname = next_name(dimname);
	}
	TRACE((stderr, "totalfieldentries is %d\n", totalfieldentries));

	/* OK, determine the actual type and size and allocate a buffer
	   for the field.
	   Apparently the only way to determine the machine type is 
	   by getting out the size (in bytes) of each element.  BLEAH!*/
	{
	  hid_t typeID;
	  H5T_class_t classID;
	  H5T_order_t order;
	  size_t tsize;

	  /* NOTE: this is an undocumented HDF-EOS API function that is
	     apparently the only way to get the real type information out.
	  */
	  SS(HE5_GDinqdatatype(gdhandle, fieldname, NULL, HE5_HDFE_DATAGROUP,
			       &typeID, &classID, &order, &tsize));
	  TRACE((stderr, "tsize is %d\n", tsize));

	  databuf = (void *)malloc(totalfieldentries * tsize);
	  if(databuf == NULL) {
	    /* Yes, I could write arbitrarily hairy code to read/write the
	       field one slice at a time, but I'm feeling a little lazy today...
	    */
	    fprintf(stderr, "Can't convert grid %s, field %s of type %d - can't allocate enough memory\n", gdname, fieldname, fieldtype);
	    /* NOTE again - skipping to the next field here */
	    fieldname = next_name(fieldname);
	    continue;
	  }

	  /* Here we map the HDF type to a netCDF type using the size
	     information we just extracted */
	  switch(classID) {
	  case H5T_INTEGER:
	    switch(tsize) {
	    case 1: nctype = NC_BYTE; break; /* NC_CHAR? */
	    case 2: nctype = NC_SHORT; break;
	    case 4: nctype = NC_INT; break;
	    default: spit_status(-1, "Bogus size for integer number");
	    }
	    break;
	  case H5T_FLOAT:
	    switch(tsize) {
	    case 4: nctype = NC_FLOAT; break;
	    case 8: nctype = NC_DOUBLE; break;
	    default: spit_status(-1, "Bogus size for floating-point number");
	    }
	    break;
	  default:
	    /* OK, let's explicitly throw this one away and just warn */
	    fprintf(stderr, "Can't convert grid %s, field %s of type %d\n",
		    gdname, fieldname, fieldtype);
	    /* NOTE: skipping to the next field here */
	    fieldname = next_name(fieldname);
	    continue;
	  }
	}
	/* Apparently the documentation that says all the start/stride/end
	   fields can be defaulted is lying.  We need to send it sane arrays
	*/
	SS(HE5_GDreadfield(gdhandle, fieldname, startdim, NULL, dimsizebuf, databuf));

	/* OK we've theoretically read the data in, now hose it back out */
	/* We construct variable names in the result file as
	   gridname_fieldname, to insure they're globally unique.
	   We may have to translate variable names here - netCDF variable
	   names are pretty restrictive ( [a-zA-Z][0-9a-zA-Z_]* )
	*/
	varname = newstring(strlen(fieldname)+strlen(gdname) + 2);
	sprintf(varname, "%s_%s", gdname, fieldname);
	fix_nc_varname(varname);

	/* netCDF files are moded, so for the convenience of this loop
	   you get to switch into and out of define mode.  Bleah. */
	/* I hope this isn't a  major performance drag - if it is, I'll have
	   to restructure this loop.  Bleah**2 */
	SS(nc_redef(ncid));

	SS(nc_def_var(ncid, varname, nctype, numdim, dimidbuf, &varid));
	/* if there are any attributes on this field, move them to this
	   variable now while we're still in define mode */
	SS(nc_enddef(ncid));
	switch(nctype) {
	case NC_BYTE:
	  SS(nc_put_var_uchar(ncid, varid, (unsigned char *)databuf));
	  break;
	case NC_SHORT:
	  SS(nc_put_var_short(ncid, varid, (short *)databuf));
	  break;
	case NC_INT:
	  SS(nc_put_var_int(ncid, varid, (int *)databuf));
	  break;
	case NC_FLOAT:
	  SS(nc_put_var_float(ncid, varid, (float *)databuf));
	  break;
	case NC_DOUBLE:
	  SS(nc_put_var_double(ncid, varid, (double *)databuf));
	  break;
	default:
	  fprintf(stderr, "Grid %s, field %s, netCDF type %d, ???\n",
		  gdname, fieldname, nctype);
	  spit_status(-1, "How the hell did we get here???");
	  break;
	}
	/* If we have variable/grid local attributes, dump them here */
	TRACE((stderr, "ZONECODE is %d\n", zonecode));
	if(zonecode > 0){
	  char *zonestring = utmzonecodemap[zonecode];
	  nc_redef(ncid);
	  SS(nc_put_att_text(ncid, varid, "UTMZone", strlen(zonestring), zonestring));
	  TRACE((stderr, "Variable %d Attribute |%s|: |%s|\n",
		 varid, "UTMZone", zonestring));
	}
	TRACE((stderr, "PROJCODE is %d\n", projcode));
	if(projcode >= 0) {
	  struct proj_info *pi = find_GPI(projcode);
	  TRACE((stderr, "GPI object is %x\n", pi));
	  if(pi != NULL) {
	    int i;
	    nc_redef(ncid);
	    SS(nc_put_att_text(ncid, varid, "GCTP_Projection_Name", strlen(pi->projection_name), pi->projection_name));
	    TRACE((stderr, "Variable %d Attribute |%s|: |%s|\n",
		   varid, "GCTP_Projection_Name", pi->projection_name));
	    TRACE((stderr, "Moving %d projection parameters\n", pi->tagcount));
	    for (i=0; i < pi->tagcount; i++) {
	      char *parmtag = pi->tags[i].name;
	      SS(nc_put_att_double(ncid, varid, parmtag, NC_DOUBLE, 1, &(projparm[pi->tags[i].index])));
	      TRACE((stderr, "Variable %d Attribute |%s|: %f\n",
		     varid, parmtag, projparm[pi->tags[i].index]));
	    }
	  }
	}
	if(spherecode >= 0) {
	  char *sphname = spheroid_names[spherecode];
	  SS(nc_put_att_text(ncid, varid, "GCTP_Spheroid_Name", strlen(sphname), sphname));
	  TRACE((stderr, "Variable %d Attribute |%s|: |%s|\n",
		 varid, "GCTP_Spheroid_Name", sphname));
	}
	      
	free(varname);
	free(databuf);
	fieldname = next_name(fieldname);
      }

      /* free all the junk we just malloc-ed */
      free(fieldnamebuf);
    }
    /* move the attributes */
    move_grid_attributes(gdhandle, gdname, ncid);
    free(dimsizebuf);
    free(dimnamebuf);
    HE5_GDdetach(gdhandle);
    gdname = next_name(gdname);
  }
  free(gdnamebuf);
  HE5_GDclose(filehandle);
  nc_close(ncid);
}

/* build a coordinate vector for a dimension entry */
void make_grid_cv(int ncid, struct dment *de,
		  double start, double end, int convertDMS)
{
  /* A coordinate vector is a one-dimensional vector with the
     same name as the dimension, mapping the integer dimension index
     to some other space.  In the grid case, we map it to evenly-spaced
     points such that 0 == start and (de->size -1) == end.

     If convertDMS is on, we use the HDF-EOS converter to change its
     encoded DDDMMMSSS.SS to nice, flat, math-friendly decimal degrees.
  */

  double *dvec = (double *) malloc(de->size * sizeof(double));
  double grid_spacing;
  int dimarray[1] = {de->id};
  int varid;
  int i;

  if(convertDMS) {
    start = HE5_EHconvAng(start, HE5_HDFE_DMS_DEG);
    end = HE5_EHconvAng(end, HE5_HDFE_DMS_DEG);
  }

  grid_spacing = (end - start) / de->size;

  for(i=0; i<de->size; i++) {
    dvec[i] = start + (i * grid_spacing);
  }
  TRACE((stderr, "Defining coordinate vector |%s| of size %d from %f to %f\n",
	 de->Nname, de->size, start, end));
  nc_redef(ncid);
  SS(nc_def_var(ncid, de->Nname, NC_DOUBLE, 1, dimarray, &varid));
  nc_enddef(ncid);
  SS(nc_put_var_double(ncid, varid, dvec));
  free(dvec);
}


/* this moves grid attributes to netCDF global attributes */
/* In addition to the inqattrs/readattr pair, there also exist
   inqgrpattrs/readgrpattr and inqlocattrs/readlocattr.  Problem is,
   there is only attrinfo, which you must have to get the real machine-local
   data type of attributes.  So, I'm going to punt on converting the
   grp and loc attributes.  Foo!
*/
void move_grid_attributes(int grhandle, char *grname, int ncid)
{
  char *attnamebuf;
  long attbufsize;
  char *attname;
  int attcount;
  int attindex;
  int nctype;
  /* just long enough for the simple scalars */
  char tempatt[9];

  /* we need to be in netCDF define mode here */
  nc_redef(ncid);

  SS(attcount = HE5_GDinqattrs(grhandle, NULL, &attbufsize));
  attnamebuf = newstring(attbufsize+1);
  SS(HE5_GDinqattrs(grhandle, attnamebuf, &attbufsize));
  
  TRACE((stderr, "Moving %d attributes of grid |%s|:\n\t|%s|\n",
	 attcount, grname, attnamebuf));
  attname = attnamebuf;
  for(attindex = 0; attindex < attcount; attindex++) {
    H5T_class_t atttype;
    int attsize;
    char *ncattname;
    void *attentbuf;
    hsize_t atteltcount;

    int tsize;

    end_name(attname);
    ncattname = newstring(strlen(grname)+strlen(attname)+2);
    sprintf(ncattname, "%s_%s", grname, attname);
    fix_nc_varname(ncattname);
    TRACE((stderr, "Attribute |%s| moves to |%s|\n", attname, ncattname));

    SS(HE5_GDattrinfo(grhandle, attname, &atttype, &atteltcount));
    TRACE((stderr, "Type number %d, size %d\n", atttype, atteltcount));

    { 
      hid_t typeID;
      H5T_class_t classID;
      H5T_order_t order;
      /* NOTE: this is an undocumented HDF-EOS API function that is
	 apparently the only way to get the real type information out.
      */
      SS(HE5_GDinqdatatype(grhandle, NULL, attname, HE5_HDFE_ATTRGROUP,
			   &typeID, &classID, &order, &tsize));
    }

    /* We have the number of attribute elements, their size, and their
       abstract type, which is what we need to map them to their
       netCDF concrete type.
    */
    switch(atttype) {
    case H5T_INTEGER:
      switch(tsize) {
      case 1: nctype = NC_BYTE; break;
      case 2: nctype = NC_SHORT; break;
      case 4: nctype = NC_INT; break;
      default: spit_status(-1, "Bogus size for integer number");
      }
      break;
    case H5T_FLOAT:
      switch(tsize) {
      case 4: nctype = NC_FLOAT; break;
      case 8: nctype = NC_DOUBLE; break;
      default: spit_status(-1, "Bogus size for floating-point number");
      }
      break;
    case H5T_STRING:
      /* Strings are a weird special case in HDF-EOS, as they are stored
	 not as vectors of char, but as scalars with a specific size, so
	 here, tsize IS the number of characters.  To make the rest of the
	 code work, we smack it to 1 here.
      */
      nctype = NC_CHAR;
      atteltcount = tsize;
      tsize = 1;
      break;
    default:
      fprintf(stderr, "Unknown type of attribute: %d\n", atttype);
      spit_status(-1, "Unhandled attribute type");
      break;
    }
    attsize = atteltcount * tsize;

    if(attsize < 9)
      attentbuf = (void *)tempatt;
    else
      attentbuf = newstring(attsize);
    SS(HE5_GDreadattr(grhandle, attname, attentbuf));

    switch(nctype) {
    case NC_BYTE:
      SS(nc_put_att_uchar(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (char *)attentbuf));
      break;
    case NC_CHAR:
      SS(nc_put_att_text(ncid, NC_GLOBAL, ncattname, atteltcount, (char *)attentbuf));
      break;
    case NC_SHORT:
      SS(nc_put_att_short(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (short *)attentbuf));
      break;
    case NC_INT:
      SS(nc_put_att_int(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (int *)attentbuf));
      break;
    case NC_FLOAT:
      SS(nc_put_att_float(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (float *)attentbuf));
      break;
    case NC_DOUBLE:
      SS(nc_put_att_double(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (double *)attentbuf));
      break;
    }
    /* Free the attribute memory */
    if(attentbuf != tempatt)
      free(attentbuf);
    attname = next_name(attname);
  }
}
